{ MERGE.PAS : Merge ordered ASCII textfiles ascending/descending

  Title   : MERGE
  Language: Borland Pascal v5.0 or higher
  Version : 1.4
  Date    : Feb 15, 2000
  Author  : J R Ferguson
  Usage   : refer procedure Help
  Download: http://hello.to/ferguson
  E-mail  : j.r.ferguson@iname.com

This program and its source may be used and copied freely without charge,
but  only  for non-commercial purposes. The author is not responsible for
any damage or loss of data that may be caused by using it.

To compile this source file, you wil need  some  units  from  the  JRFPAS
Pascal  routine  library by the same author, which can be downloaded from
the Internet address mentioned above.
}

{$V-}

program merge;

uses DefLib, ArgLib, CvtLib, SckLib, StpLib, ChrLib;

const
  PROGIDN   = 'MERGE';
  PROGVERS  = 'v1.3';
  MAXFNAME  = 79;
  DFLMSG    = 'CON';

  ERROK     = 0;
  ERRARG    = 1;
  ERRINP    = 2;
  ERROUT    = 3;
  ERRIND    = 4;
  ERRMAX    = 4;

  ERRMSG    : array[1..ERRMAX] of StpTyp =
 (''
 ,'File not found : '
 ,'Can''t open output : '
 ,'key index out of range (must be 1..255)'
 );

type
  InpFrec   = record
		fvar : text;
                line : StpTyp;
              end;
  OLLelm    = InpFrec;
  OLLptr    = ^OLLentry;
  OLLentry  = record
                inh : OLLelm;
                nxt : OLLptr
              end;

var
  InpFname,
  OutFname  : StpTyp;
  MsgFvar,
  OutFvar   : text;
  OutOpen   : boolean;
  InpFlist  : OLLptr;
  InpFbuff  : OLLelm;

  ErrCod    : integer;
  ErrFname  : StpTyp;

  KeyPos,
  KeyLen    : StpInd;
  OptHlp,
  OptDes,
  OptIgn,
  OptLex    : boolean;


{--- General ---}

procedure Help;
  procedure wi(i: integer); begin write  (MsgFvar,i); end;
  procedure wr(s: StpTyp ); begin write  (MsgFvar,s); end;
  procedure wl(s: StpTyp ); begin writeln(MsgFvar,s); end;
begin
wl(PROGIDN+' '+PROGVERS+' - Merge ordered ASCII textfiles');
wl('Usage  : '+PROGIDN+' [<infile1] infile2 [...] [>outfile] [/option[...] [...]]');
wl('Options:');
wr('    K[c1][,c2]   Define key first column c1 [1], last column c2 [');
        wi(MaxStp); wl('].');
wr(' or K[c1][,+c2]  Define key first column c1 [1], nbr of cols c2 [');
        wi(MaxStp); wl('].');
wl('    D  Sort in descending order.');
wl('    I  Ignore upper/lower case. Implied by /L.');
wl('    L  Sort in lexical order ' +
          '(control,punctuation,digits,letters). Implies /I.');
wl('    H  Send this help text to (redirected) output.');
end;

function CompareResult(str1,str2: StpTyp): integer;
var tmp1, tmp2: StpTyp; result: integer;
begin
  StpSub(tmp1,str1,KeyPos,KeyLen);
  StpSub(tmp2,str2,KeyPos,KeyLen);
  if OptLex then result:= StpLexCmp(tmp1,tmp2)
  else begin
    if OptIgn then begin StpUpp(tmp1); StpUpp(tmp2) end;
    result:= StpCmp(tmp1,tmp2);
  end;
  if OptDes then CompareResult:= - result else CompareResult:= result;
end;


{--- Ordered Linked List routines ---}

function OLLorder(e1, e2: OLLelm): integer;
{ result < 0 if e1 < e2 , 0 if e1 = e2 , > 0 if e1 > e2 }
begin OLLorder:= CompareResult(e1.line,e2.line) end;

procedure OLLcreate(var oll: OLLptr);
  begin new(oll); oll^.nxt:= oll end;

procedure OLLclear(var oll: OLLptr);
begin dispose(oll) end;

function OLLempty(var oll: OLLptr): boolean;
  begin OLLempty:= oll^.nxt=oll end;

procedure OLLinsert(var elm: OLLelm; var oll: OLLptr);
  var pt0,pt1   : OLLptr;
  begin
    oll^.inh:= elm; {plaats sentinel}
    pt0:= oll;
    while OLLorder(pt0^.nxt^.inh,elm) < 0 do pt0:= pt0^.nxt;
    new(pt1); with pt1^ do begin inh:= elm; nxt:= pt0^.nxt end;
    pt0^.nxt:= pt1
  end;

procedure OLLretrieve(var elm: OLLelm; var oll: OLLptr);
  var ptr: OLLptr;
  begin
    ptr:= oll^.nxt;
    elm:= ptr^.inh;
    oll^.nxt:= ptr^.nxt;
    if ptr^.nxt <> ptr then dispose(ptr)
  end;


{--- I/O routines ---}

procedure OpenMsg;
begin
  if OptHlp then Assign(MsgFVar,OutFname) else Assign(MsgFVar,DFLMSG);
  rewrite(MsgFVar);
end;

procedure CloseMsg;
begin Close(MsgFVar) end;

procedure InpLine;
begin with InpFbuff do begin
  if eof(fvar) then Close(fvar)
  else begin readln(fvar,line); OLLinsert(InpFbuff,InpFlist) end;
end end;

procedure OpenInp(fname: StpTyp);
begin with InpFbuff do begin
  {$I-} Assign(fvar,fname); reset(fvar) {$I+};
  if IOresult=0 then InpLine else begin ErrCod:= ERRINP; ErrFname:= fname end;
end end;

procedure OpenOut;
begin
  {$I-} Assign(OutFvar,OutFname); rewrite(OutFvar) {$I+};
  if IOresult <> 0 then begin ErrCod:= ERROUT; ErrFname:= OutFname end
  else OutOpen:= true
end;


{--- Main line ---}

function ReadUns(var arg: StpTyp): integer;
var n: integer; c: char;
begin
  n:= 0; c:= StpcRet(arg,1);
  while (ErrCod=ERROK) and IsDigit(c) do begin
    StpDel(arg,1,1);
    n:= 10 * n + (ord(c) - ord('0'));
    c:= StpcRet(arg,1);
  end;
  ReadUns:= n;
end;

procedure ReadKeyDesc(var arg: StpTyp);
var n: integer;
begin
  if IsDigit(StpcRet(arg,1)) then KeyPos:= ReadUns(arg) mod MaxStp;
  if StpcRet(arg,1) = ',' then begin
    StpDel(arg,1,1);
    if StpcRet(arg,1) = '+' then begin
      StpDel(arg,1,1);
      if IsDigit(StpcRet(arg,1)) then KeyLen:= ReadUns(arg) mod MaxStp
      else ErrCod:= ERRARG;
    end
    else begin
      if IsDigit(StpcRet(arg,1)) then begin
        n:= (ReadUns(arg) mod MaxStp) - KeyPos + 1;
        if n < 0 then ErrCod:= ERRARG else KeyLen:= n;
      end
      else ErrCod:= ERRARG;
    end;
  end;
end;

procedure ReadOpt(arg: StpTyp);
var nextopt: boolean;
begin
  StpDel(arg,1,1);
  repeat
    if StpEmpty(arg) or (StpcRet(arg,1) = '/') then ErrCod:= ERRARG
    else begin
      nextopt:= false;
      while (ErrCod=ERROK) and not nextopt and not StpEmpty(arg) do
      case StpCGet(arg) of
       'D': OptDes := true;
       'I': OptIgn := true;
       'H': OptHlp := true;
       'L': OptLex := true;
       'K': ReadKeyDesc(arg);
       '/': nextopt:= true;
       else ErrCod:= ERRARG;
      end;
    end;
  until (ErrCod <> ERROK) or not nextopt;
end;

procedure ReadArgs;
var i   : ArgInd;
    arg : StpTyp;
    Fcnt: integer;
begin
  GetArgs;
  i:= 0; Fcnt:= 1;
  while (i < ArgC) and (ErrCod = ERROK) do begin
    Inc(i); StpCpy(arg,ArgV[i]); StpUpp(arg);
    case StpCRet(arg,1) of
      '/' : ReadOpt(arg);
      '<' : StpSub(InpFname,arg,2,MAXFNAME);
      '>' : StpSub(OutFname,arg,2,MAXFNAME);
      else  begin Inc(Fcnt); StpSub(arg,arg,1,MAXFNAME); OpenInp(arg) end;
    end;
  end;
  if OptHlp then if ErrCod = ERROK then ErrCod:= ERRARG else OptHlp:= false;
  if ErrCod = ERROK then if Fcnt < 2 then ErrCod:= ERRARG;
end;

procedure MainInit;
begin
  KeyPos:= 1; KeyLen:= MAXSTP;
  OptHlp:= false; OptDes:= false; OptIgn:= false; OptLex:= false;
  StpCreate(InpFname); StpCreate(OutFname); { Standard input/output }
  ErrCod:= ERROK; OutOpen:= false;
  OLLcreate(InpFlist);
  ReadArgs;
  if ErrCod = ERROK then begin
    OpenInp(InpFname);
    if ErrCod = ERROK then OpenOut;
  end;
end;

procedure MainTerm;
begin
  OLLclear(InpFlist);
  if OutOpen then Close(OutFvar);
  if ErrCod <> ERROK then begin
    OpenMsg;
    if ErrCod=ERRARG then Help
    else begin
      write(MsgFvar,'MERGE: ',ERRMSG[ErrCod]);
      if (ErrCod = ERRINP) or (ErrCod = ERROUT) then write(MsgFvar,ErrFname);
    end;
    writeln(MsgFvar);
    CloseMsg;
  end;
end;

begin { Main }
  MainInit;
  if ErrCod = ERROK then
    while not OLLempty(InpFlist) do begin
      OLLretrieve(InpFbuff,InpFlist);
      writeln(OutFvar,InpFbuff.line);
      InpLine;
    end;
  MainTerm;
end.
